/*
GFCLK_SYS.CPP
Nov 2004 saw@SFB7
*/

#include <ntddk.h>
#include <string.h>

#include "gfclk_sys.h"


#define NT_DEVICE_NAME		L"\\Device\\GFCLK"
#define DOS_DEVICE_NAME 	L"\\DosDevices\\GFClk"

NTSTATUS DriverEntry(
	IN OUT PDRIVER_OBJECT	DriverObject,
	IN PUNICODE_STRING		RegistryPath
	);

NTSTATUS GFClkCreateClose(
	IN PDEVICE_OBJECT		DeviceObject,
	IN PIRP 				Irp
	);

NTSTATUS GFClkDeviceControl(
	IN PDEVICE_OBJECT		DeviceObject,
	IN PIRP 				Irp
	);

VOID GFClkUnloadDriver(
	IN PDRIVER_OBJECT		DriverObject
	);
	

#ifdef ALLOC_PRAGMA
#pragma alloc_text( INIT, DriverEntry )
#pragma alloc_text( PAGE, GFClkCreateClose)
#pragma alloc_text( PAGE, GFClkDeviceControl)
#pragma alloc_text( PAGE, GFClkUnloadDriver)
#endif // ALLOC_PRAGMA


NTSTATUS DriverEntry(
	IN OUT PDRIVER_OBJECT	DriverObject,
	IN PUNICODE_STRING		RegistryPath
	)
{
	NTSTATUS		ntStatus;
	UNICODE_STRING	ntUnicodeString;
	UNICODE_STRING	ntWin32NameString;
	PDEVICE_OBJECT	deviceObject = NULL;

	RtlInitUnicodeString( &ntUnicodeString, NT_DEVICE_NAME );
	
	ntStatus = IoCreateDevice(
		DriverObject,					// Our Driver Object
		0,								// We don't use a device extension
		&ntUnicodeString,				// Device name "\Device\GFCLK"
		FILE_DEVICE_UNKNOWN,			// Device type
		FILE_DEVICE_SECURE_OPEN,	 // Device characteristics
		FALSE,							// Not an exclusive device
		&deviceObject );				// Returned ptr to Device Object

	if (!NT_SUCCESS(ntStatus)) return ntStatus;
	
	//
	// Initialize the driver object with this driver's entry points.
	//

	DriverObject->MajorFunction[IRP_MJ_CREATE] = GFClkCreateClose;
	DriverObject->MajorFunction[IRP_MJ_CLOSE] = GFClkCreateClose;
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = GFClkDeviceControl;
	DriverObject->DriverUnload = GFClkUnloadDriver;
   
	//
	// Initialize a Unicode String containing the Win32 name
	// for our device.
	//

	RtlInitUnicodeString( &ntWin32NameString, DOS_DEVICE_NAME );

	//
	// Create a symbolic link between our device name  and the Win32 name 
	//

	ntStatus = IoCreateSymbolicLink(
						&ntWin32NameString, &ntUnicodeString );

	if (!NT_SUCCESS(ntStatus)) IoDeleteDevice(deviceObject);

	return ntStatus;
}


NTSTATUS GFClkCreateClose(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
	Irp->IoStatus.Status = STATUS_SUCCESS;
	Irp->IoStatus.Information = 0;
	
	IoCompleteRequest( Irp, IO_NO_INCREMENT );
	
	return STATUS_SUCCESS;
}

VOID GFClkUnloadDriver(IN PDRIVER_OBJECT DriverObject)
{
	PDEVICE_OBJECT deviceObject = DriverObject->DeviceObject;
	UNICODE_STRING uniWin32NameString;

	//
	// Create counted string version of our Win32 device name.
	//

	RtlInitUnicodeString( &uniWin32NameString, DOS_DEVICE_NAME );


	//
	// Delete the link from our device name to a name in the Win32 namespace.
	//
	
	IoDeleteSymbolicLink( &uniWin32NameString );
	
	if ( deviceObject != NULL )
	{
		IoDeleteDevice( deviceObject );
	}

 

}

NTSTATUS GFClkDeviceControl(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
	PIO_STACK_LOCATION	irpSp;// Pointer to current stack location
	NTSTATUS			ntStatus = STATUS_SUCCESS;// Assume success
	ULONG				inBufLength, outBufLength;
	void				*outBuf; // pointer to Input and output buffer
	PHYSICALRDWR		*rdwrparam;
	PHYSICAL_ADDRESS	address;
	void				*mappedmem;
	PCICONFIGRDWR		*pciconfparam;
	
	irpSp = IoGetCurrentIrpStackLocation( Irp );
	inBufLength = irpSp->Parameters.DeviceIoControl.InputBufferLength;
	outBufLength = irpSp->Parameters.DeviceIoControl.OutputBufferLength;

	switch (irpSp->Parameters.DeviceIoControl.IoControlCode)
	{
	case GFCLK_READ_MEM:
		rdwrparam = (PHYSICALRDWR *)Irp->AssociatedIrp.SystemBuffer;
		outBuf = Irp->AssociatedIrp.SystemBuffer;
		address.HighPart = 0;
		address.LowPart = (ULONG)rdwrparam->address;
		mappedmem = MmMapIoSpace(address, 4, FALSE);
		Irp->IoStatus.Information = 0;
		switch(rdwrparam->size){
		case 1:
			READ_REGISTER_BUFFER_UCHAR((PUCHAR)mappedmem, (PUCHAR)outBuf, 1);
			Irp->IoStatus.Information = 1;
			break;
		case 2:
			READ_REGISTER_BUFFER_USHORT((PUSHORT)mappedmem, (PUSHORT)outBuf, 1);
			Irp->IoStatus.Information = 2;
			break;
		case 4:
			READ_REGISTER_BUFFER_ULONG((PULONG)mappedmem, (PULONG)outBuf, 1);
			Irp->IoStatus.Information = 4;
			break;
		}
		MmUnmapIoSpace(mappedmem, 4);
		break;
		
	case GFCLK_WRITE_MEM:
		rdwrparam = (PHYSICALRDWR *)Irp->AssociatedIrp.SystemBuffer;
		address.HighPart = 0;
		address.LowPart = (ULONG)rdwrparam->address;
		mappedmem = MmMapIoSpace(address, 4, FALSE);
		Irp->IoStatus.Information = 0;
		switch(rdwrparam->size){
		case 1:
			WRITE_REGISTER_BUFFER_UCHAR((PUCHAR)mappedmem, (PUCHAR)&rdwrparam->data, 1);
			break;
		case 2:
			WRITE_REGISTER_BUFFER_USHORT((PUSHORT)mappedmem, (PUSHORT)&rdwrparam->data, 1);
			break;
		case 4:
			WRITE_REGISTER_BUFFER_ULONG((PULONG)mappedmem, (PULONG)&rdwrparam->data, 1);
			break;
		}
		MmUnmapIoSpace(mappedmem, 4);
		break;
	case GFCLK_READ_PCICONF:
		pciconfparam = (PCICONFIGRDWR *)Irp->AssociatedIrp.SystemBuffer;
		outBuf = Irp->AssociatedIrp.SystemBuffer;
		if(HalGetBusData(PCIConfiguration, pciconfparam->bus, pciconfparam->slot,
			outBuf, pciconfparam->length) != pciconfparam->length){
			Irp->IoStatus.Information = 0;
			ntStatus = STATUS_INVALID_PARAMETER;
		}
		else
			Irp->IoStatus.Information = pciconfparam->length;
		break;
		
	default:

		ntStatus = STATUS_INVALID_DEVICE_REQUEST;
		break;
	}

	Irp->IoStatus.Status = ntStatus;

	IoCompleteRequest(Irp, IO_NO_INCREMENT);

	return ntStatus;
}
